-module(mup_bugtracking).
-behaviour(gen_server).
-export([start_link/1]).
-export([init/1, handle_call/3, handle_cast/2, handle_info/2,
         terminate/2, code_change/3]).
-include("mup.hrl").
-include_lib("eunit/include/eunit.hrl").


-record(state,
        {plugin,
         port,
         bugs,
         update_timer}).


start_link(Name) ->
    gen_server:start_link(?MODULE, Name, []).


init(Name) ->
    {ok, Plugin} = mup_config:get_plugin(Name),
    ScrapBugsPath = filename:join(code:priv_dir(mup), "scrapbugs.py"),
    Port = open_port({spawn, ScrapBugsPath}, [{line, 65536}]),
    {ok, schedule_update(#state{plugin=Plugin, port=Port})}.

handle_call(_Request, _From, State) ->
    {reply, {error, bad_request}, State}.

handle_cast({handle_message, _Message}, State) ->
    {noreply, State};

handle_cast(reconfig, State) ->
    case mup_config:get_plugin((State#state.plugin)#plugin.name) of
        {ok, Plugin} ->
            {noreply, State#state{plugin=Plugin}};
        undefined ->
            {stop, normal, State}
    end.

handle_info(update, State) ->
    LaunchpadPid = get_mup_launchpad_pid(State),
    Url = get_option(url, State),
    Port = State#state.port,
    port_command(Port, Url ++ "\n"),
    NewBugs = ordsets:from_list(string:tokens(receive_line(Port), " ")),
    case State#state.bugs of
        undefined ->
            ok;
        OldBugs ->
            Channels = get_option(channels, [], State),
            case get_option(prefix_for_old, State) of
                undefined ->
                    ok;
                PrefixForOld ->
                    show_bugs(ordsets:subtract(OldBugs, NewBugs), Channels,
                              PrefixForOld, LaunchpadPid)
            end,
            case get_option(prefix_for_new, State) of
                undefined ->
                    ok;
                PrefixForNew ->
                    show_bugs(ordsets:subtract(NewBugs, OldBugs), Channels,
                              PrefixForNew, LaunchpadPid)
            end
    end,
    {noreply, schedule_update(State#state{bugs=NewBugs})};

handle_info(_Info, State) ->
    {noreply, State}.

terminate(_Reason, State) ->
    port_close(State#state.port),
    ok.

code_change(_OldVsn, State, _Extra) ->
    {ok, State}.


get_option(Key, State) ->
    mup_util:option(Key, (State#state.plugin)#plugin.options).

get_option(Key, Default, State) ->
    mup_util:option(Key, (State#state.plugin)#plugin.options, Default).

get_mup_launchpad_pid(State) ->
    case get_option(mup_launchpad_name, State) of
        undefined ->
            throw({missing_option, mup_launchpad_name});
        Name ->
            {ok, LaunchpadPid} = mup_plugin_sup:find_plugin(Name),
            LaunchpadPid
    end.


schedule_update(State) ->
    case State#state.update_timer of
        undefined ->
            ok;
        Timer ->
            erlang:cancel_timer(Timer)
    end,
    Interval = get_option(interval, 300, State) * 1000,
    State#state{update_timer=erlang:send_after(Interval, self(), update)}.


receive_line(Port) ->
    receive_line(Port, []).

receive_line(Port, Parts) ->
    receive
        {Port, {data, {noeol, Part}}} ->
            receive_line(Port, [Part|Parts]);
        {Port, {data, {eol, Part}}} ->
            lists:flatten(lists:reverse([Part|Parts]))
    after
        15000 ->
            throw(port_timeout)
    end.


show_bugs([Bug|Rest], Channels, Prefix, LaunchpadPid) ->
    show_bug(Bug, Channels, Prefix, LaunchpadPid),
    show_bugs(Rest, Channels, Prefix, LaunchpadPid);

show_bugs([], _Channels, _Prefix, _LaunchpadPid) ->
    ok.

show_bug(Bug, [{Server, Channel}|Rest], Prefix, LaunchpadPid) ->
    mup_launchpad:show_bug(LaunchpadPid, Prefix, list_to_integer(Bug),
                           Server, Channel),
    show_bug(Bug, Rest, Prefix, LaunchpadPid);

show_bug(_Bug, [], _Prefix, _LaunchpadPid) ->
    ok.
